ax.extension.codemirror = {}
ax.extension.codemirror.report = {}
ax.extension.codemirror.form = {}
ax.extension.codemirror.form.control = function(
  f, options = {}
) {

  let a = ax.a

  return a['|appkit-form-control'](
    a['|appkit-form-codemirror']([
      this.control.toolbar(f, options),
      this.control.editor(f, options),
    ], {
      $setMode: function() {
        this.$('textarea').$codemirror.setOption(
          'mode',
          this.$('|appkit-form-codemirror-mode').$value()
        )
      },
      $setKeymap: function() {
        this.$('textarea').$codemirror.setOption(
          'keyMap',
          this.$('|appkit-form-codemirror-keymap').$value() || null,
        )
      },
    }), {
      $value: function() {
        return this.$('textarea').$codemirror.getValue()
      },
      $focus: function() {
        this.$('textarea').$codemirror.focus()
      },
      $disable: function() {
        this.$$('textarea').setAttribute('disabled', 'disabled')
      },
      $enable: function() {
        this.$('textarea').$refresh()
        if (!options.disabled) {
          this.$$('textarea').removeAttribute('disabled')
        }
      },

      ...options.controlTag

    }
  )

}
ax.extension.codemirror.report.control = function(
  r, options = {}
) {

  let a = ax.a
  let x = ax.x
  let report = x.codemirror.report

  return a['|appkit-report-control'](

    a['|appkit-report-codemirror'](
      [
        report.control.toolbar(options),
        report.control.editor(options),
      ], {
        ...options.codeTag
      }
    ),

    {

      'data-name': options.name,
      $value: function() {
        return options.value
      },

      tabindex: 0,
      $focus: function() {
        this.focus()
      },

      ...options.controlTag

    }

  )

}
ax.extension.codemirror.form.control.editor = function(
  f, options = {}
) {

  let a = ax.a

  return a['|appkit-form-codemirror-editor'](
    f.textarea({
      name: options.name,
      value: options.value,
      textareaTag: {
        $init: function() {
          this.$setup()
        },
        $refresh: function() {
          setTimeout(function() {
            this.$codemirror.refresh()
          }.bind(this), 1)
        },
        $setup: function() {

          this.$codemirror = CodeMirror.fromTextArea(this, {
            lineNumbers: true,
            placeholder: options.placeholder,
          })

          this.$('^|appkit-form-codemirror').$setMode()

          this.$codemirror.setSize('100%', '100%')

          this.$required()
          this.$refresh()

          this.$codemirror.on('keyup', function(codemirror, e) {
            codemirror.getTextArea().$required()

          })

        },
        $required: function() {
          let value = this.$codemirror.getValue()
          let textarea = this.$('^|appkit-form-codemirror').$$('textarea')[1]
          if (!value && options.required) {
            textarea.setAttribute('required', true)
          } else {
            textarea.removeAttribute('required')
          }
        },
        ...options.textareaTag
      },
    }), {
      $on: {
        'keyup: send control change event': (e, el) =>
          el.$send('ax.appkit.form.control.change'),
        'keyup: update textarea value': function(e) {
          this.$('textarea').$codemirror.save()
        },
        'keydown: check for editor exit': function(e) {

          let container = this.$('^|appkit-form-codemirror')
          let allowEsc = this.$('textarea').$codemirror.options.keyMap != 'vim'

          if (container.classList.contains('fullscreen')) {

            if (e.keyCode == 27 && (allowEsc || e.ctrlKey)) {
              // ESC pressed - close full screen
              container.$('|appkit-form-codemirror-fullscreen button').click()
            }

          } else {

            if (e.keyCode == 27 && e.shiftKey) {
              // shift+ESC pressed - move focus backward
              ax.x.lib.tabable.previous(e.target).focus()
            } else if (e.keyCode == 27 && (allowEsc || e.ctrlKey)) {
              // ESC pressed - move focus forward
              ax.x.lib.tabable.next(e.target).focus()
            }

          }

        }
      }
    }
  )

}
ax.extension.codemirror.form.control.keymap = function(
  f, options
) {

  let a = ax.a
  let x = ax.x

  let keymap
  let component
  let value

  let keymapLabel = function(keymap) {
    let labels = {
      vim: 'Vim',
      emacs: 'Emacs',
      sublime: 'Sublime',
    }
    return labels[keymap] || keymap
  }

  if (ax.is.string(options)) {

    component = a.label(keymapLabel(options))
    value = () => options

  } else if (options) {

    if (ax.is.not.object(options)) {
      options = {}
    }

    let show
    let selections = options.selections

    if (ax.is.undefined(selections)) {

      selections = [
        ['default', '𝍣 Keys'],
      ]

      if (CodeMirror.keyMap.vim) {
        show = true
        selections.push(['vim', 'Vim'])
      }
      if (CodeMirror.keyMap.emacs) {
        show = true
        selections.push(['emacs', 'Emacs'])
      }
      if (CodeMirror.keyMap.pcSublime) {
        show = true
        selections.push(['sublime', 'Sublime'])
      }

    } else {
      show = true
    }

    if (ax.is.true(show)) {
      component = f.select({
        selections: selections,
        value: options.value,
        selectTag: {
          $on: {
            'change: set CodeMirror keymap': function() {
              this.$('^|appkit-form-codemirror').$setKeymap()
            }
          },
          ...options.selectTag
        },
      })
      value = function() {
        return this.$('select').$value()
      }
    } else {
      component = null
      value = () => 'default'
    }

  } else {

    component = null
    value = () => 'default'

  }

  return a['|appkit-form-codemirror-keymap'](component, {
    $value: value
  }, options.keymapTag)

}
ax.css({
  '|appkit-form-codemirror': {
    display: 'block',
    border: '1px solid #b3b3b3',
    '|appkit-form-codemirror-toolbar': {
      display: 'block',
      overflow: 'auto',
      color: '#333',
      backgroundColor: 'white',
      borderBottom: '1px solid #e6e6e6',
      button: {
        fontSize: '1.2em',
        border: 'none',
        backgroundColor: 'transparent',
        cursor: 'pointer'
      },
      select: {
        padding: '2px',
        border: 'none',
        backgroundColor: 'transparent',
      }
    },
    '|appkit-form-codemirror-mode, |appkit-form-codemirror-keymap': {
      'select': {
        padding: '4px',
      },
      'label': {
        margin: '2px 10px',
      }
    },
    '|appkit-form-codemirror-toolbar-right': {
      float: 'right',
    },
    '&.fullscreen': {
      position: 'fixed',
      top: 0,
      left: 0,
      right: 0,
      bottom: 0,
      border: 'none',
      borderRadius: '0px',
      zIndex: 999,
    },
    '|appkit-form-codemirror-editor': {
      'div.CodeMirror': {
        minHeight: '2em',
        border: 'unset',
        borderRadius: 'unset',
        padding: 'unset',
        fontFamily: 'monospace',
        zIndex: 1,
        'div.CodeMirror-scroll': {
          minHeight: 'unset',
        }
      }
    },
  }
})
ax.extension.codemirror.form.control.mode = function(
  f, name, options = {}
) {

  let a = ax.a

  let mode
  let component
  let value

  let selectName

  if (name.endsWith(']')) {
    selectName = name.replace(/(.*)(\])$/, "$1_mode$2")
  } else {
    selectName = name + '_mode'
  }

  if (ax.is.string(options)) {

    component = a.label(options)
    value = () => options

  } else if (options) {

    let selections = options.selections

    if (ax.is.undefined(selections)) {
      selections = Object.keys(CodeMirror.modes) // List of installed language modes
      selections.shift(); // remove null
    }

    if (ax.is.object(selections) && Object.entries(selections).length > 0) {
      component = f.select({
        name: selectName,
        placeholder: '𝍣 Mode',
        selections: selections,
        value: options.value,
        selectTag: {
          $on: {
            'change: set CodeMirror mode': function() {
              this.$('^|appkit-form-codemirror').$setMode()
            }
          },
          ...options.selectTag
        },
      })
      value = function() {
        return this.$('select').value
      }

    } else {

      component = null
      value = () => ''

    }

  } else {

    component = null
    value = () => ''

  }

  return a['|appkit-form-codemirror-mode'](component, {
    $value: value
  }, options.modeTag)

}
ax.extension.codemirror.form.control.toolbar = function(
  f, options = {}
) {

  let a = ax.a

  return a['|appkit-form-codemirror-toolbar'](
    [
      this.mode(f, options.name, options.mode || false),
      a['|appkit-form-codemirror-toolbar-right']([
        this.keymap(f, options.keymap || false),
        a['|appkit-form-codemirror-fullscreen'](
          a.button('🗖', {
            type: 'button',
            $on: {
              'click: toggle full screen': function() {
                let container = this.$('^|appkit-form-codemirror')
                let editor = container.$('|appkit-form-codemirror-editor')
                let codemirror = editor.$('textarea').$codemirror
                if (container.classList.contains('fullscreen')) {
                  this.$text = '🗖'
                  container.classList.remove('fullscreen')
                  editor.style.height = ''
                  codemirror.focus()
                } else {
                  this.$text = '🗗'
                  container.classList.add('fullscreen')
                  codemirror.focus()
                }
              }
            }
          })
        )
      ]),
    ],
    options.toolbarTag
  )

}
ax.extension.codemirror.report.control.editor = function(
  options = {}
) {

  let a = ax.a

  return a['|appkit-report-codemirror-editor'](
    a.textarea(
      options.value, {
        $init: function() {
          this.$codemirror = CodeMirror.fromTextArea(this, {
            lineNumbers: true,
            readOnly: true,
            placeholder: options.placeholder,
            mode: options.mode || '',
          })
          this.$codemirror.setSize('100%', '100%')
          setTimeout(function() {
            this.$codemirror.refresh()
          }.bind(this), 1)
        }
      }
    ), {
      $on: {
        'keydown: check for editor exit': function(e) {

          let container = this.$('^|appkit-report-codemirror')

          if (container.classList.contains('fullscreen')) {

            if (e.keyCode == 27) {
              // ESC pressed - close full screen
              container.$('|appkit-report-codemirror-fullscreen button').click()
            }

          } else {

            if (e.keyCode == 9 && e.shiftKey) {
              // shift+TAB pressed - move focus backward
              ax.x.lib.tabable.previous(e.target).focus()
            } else if (e.keyCode == 9) {
              // TAB pressed - move focus forward
              ax.x.lib.tabable.next(e.target).focus()
            }

          }

        }
      }
    }

  )

}
ax.css({
  '|appkit-report-codemirror': {
    display: 'block',
    border: '1px solid #b3b3b3',
    '|appkit-report-codemirror-toolbar': {
      display: 'block',
      overflow: 'auto',
      color: '#333',
      backgroundColor: 'white',
      borderBottom: '1px solid #e6e6e6',
      button: {
        fontSize: '1.2em',
        border: 'none',
        backgroundColor: 'transparent',
        cursor: 'pointer'
      },
      select: {
        padding: '2px',
        border: 'none',
        backgroundColor: 'transparent',
      }
    },
    '|appkit-report-codemirror-mode': {
      'select': {
        padding: '4px',
      },
      'label': {
        margin: '2px 5px',
      }
    },
    '|appkit-report-codemirror-toolbar-right': {
      float: 'right',
    },
    '&.fullscreen': {
      position: 'fixed',
      top: '0',
      left: '0',
      right: '0',
      bottom: '0',
      border: 'none',
      borderRadius: '0px',
      zIndex: '999',
    },
    '|appkit-report-codemirror-editor': {
      'div.CodeMirror': {
        minHeight: '2em',
        border: 'unset',
        borderRadius: 'unset',
        padding: 'unset',
        fontFamily: 'monospace',
        zIndex: 1,
        'div.CodeMirror-scroll': {
          minHeight: 'unset',
        }
      }
    },
  },

})
ax.extension.codemirror.report.control.toolbar = function(
  options = {}
) {

  let a = ax.a

  return a['|appkit-report-codemirror-toolbar'](
    [
      a['|appkit-report-codemirror-mode'](
        a.label(options.mode),
      ),
      a['|appkit-report-codemirror-toolbar-right'](
        a['|appkit-report-codemirror-fullscreen'](
          a.button('🗖', {
            type: 'button',
            $on: {
              'click: toggle full screen': function() {
                let container = this.$('^|appkit-report-codemirror')
                let editor = container.$('|appkit-report-codemirror-editor')
                let codemirror = editor.$('textarea').$codemirror
                if (container.classList.contains('fullscreen')) {
                  this.$text = '🗖'
                  container.classList.remove('fullscreen')
                  editor.style.height = ''
                  codemirror.focus()
                } else {
                  this.$text = '🗗'
                  container.classList.add('fullscreen')
                  codemirror.focus()
                }
              }
            }
          })
        )
      ),
    ],
  )

}
